import json
import requests
import pandas as pd
import numpy as np
from datetime import date, timedelta
from time import sleep
import os

def int_to_str(x: int) -> str:
    return "%06d" % x

def date_to_str(today):

    str_date = str(today)[:4] + str(today)[5:7] + str(today)[-2:]
    return str_date

def str_to_date(str_date):

    str_date = str_date.replace('-', '')
    year = int(str_date[:4])
    month = int(str_date[4:6])
    day = int(str_date[-2:])
    return date(year,month,day)

def get_trading_days(dir, start_date: date, end_date: date):

    df = pd.read_csv(f"{dir}/data_file/korea_trading_days.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    trading_days = df['날짜'].to_list()
    trading_days = [day for day in trading_days if start_date <= day <= end_date]
    return trading_days

def get_kosdaq_price(today:str, api):

    print(today)

    # 코스피 데이터 수집
    url = "https://data-dbg.krx.co.kr/svc/apis/idx/kosdaq_dd_trd"

    headers = {
        "AUTH_KEY": api,
        "Accept": "application/json"
    }

    params = {
        "basDd": today  # 조회할 날짜: YYYYMMDD
    }

    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()

    data = response.json()

    # OutBlock_1 추출
    records = data.get("OutBlock_1", [])

    # DataFrame 변환
    df = pd.DataFrame(records)
    df = df.replace(',', '', regex=True)
    df = df[df["IDX_NM"] == "코스닥"].reset_index(drop=True)

    df = df.drop(['IDX_CLSS', 'IDX_NM', 'CMPPREVDD_IDX', 'FLUC_RT'], axis='columns')

    df['BAS_DD'] = df['BAS_DD'].apply(lambda x : str_to_date(x))

    cols_map = {'BAS_DD':'날짜', 'CLSPRC_IDX':'종가', 'OPNPRC_IDX':'시가', 'HGPRC_IDX':'고가',
                'LWPRC_IDX':'저가', 'ACC_TRDVOL':'거래량', 'ACC_TRDVAL':'거래대금', 'MKTCAP':'시가총액'}

    df = df.rename(columns=cols_map)
    df = df[['날짜','종가','시가','고가','저가','거래량','거래대금','시가총액']]
    df = df.sort_values('날짜', ascending=True)

    return df


def generate_kosdaq_price(dir, today, api):

    end_date = today

    bef_kosdaq_price = pd.read_csv(f"{dir}/data_file/Kosdaq_Price.csv")
    bef_kosdaq_price['날짜'] = bef_kosdaq_price['날짜'].apply(lambda x : str_to_date(x))
    start_date = bef_kosdaq_price['날짜'].iloc[-1] + timedelta(days=1)

    if start_date > end_date:
        print("수집할 코스닥 INDEX가 없습니다")

    trading_days = get_trading_days(dir, start_date, end_date)

    if len(trading_days) == 0:
        print("수집할 코스닥 INDEX가 없습니다")

    else:
        trading_days = [date_to_str(day) for day in trading_days]
        aft_kosdaq_price = get_kosdaq_price(trading_days[0], api)

        for i, now_day in enumerate(trading_days):

            if i == 0:
                continue

            df2 = get_kosdaq_price(now_day, api)
            aft_kosdaq_price = pd.concat([aft_kosdaq_price, df2])
            sleep(1)

        try:
            kosdaq_price = pd.concat([bef_kosdaq_price, aft_kosdaq_price])
            kosdaq_price = kosdaq_price.reset_index()
        except:
            pass

        df = kosdaq_price[['날짜','종가','시가','고가','저가','거래량','거래대금','시가총액']]
        df.index = np.arange(len(df))

        df.to_csv(f"{dir}/data_file/Kosdaq_Price.csv")
        print("코스닥 INDEX 수집완료")



# -*- coding:utf-8 -*-
from typing import Tuple
from pandas import DataFrame, Series
import numpy as np

from buke_pypi.parameter import MovingAverage


def sma(price: Series, period: int) -> Series:
    """
    단순이동평균

    <설명>
    단순 이동평균을 구하는 함수입니다.
    단순 이동평균은 특정 기간(period) 동안의 가격(price)의 산술평균을 구합니다.

    <사용 방법>
    첫 번째 인자에는 단순 이동평균을 구하는데 사용하는 가격을,
    두 번째 인자에는 단순 이동평균을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 5일간 종가의 단순 이동평균을 구하고자 하는 경우
    'sma(close, 5)' 또는 '단순이동평균(종가, 5)'와 같이 작성하면 됩니다.

    <계산 방법>
    5일간 종가의 단순 이동평균은 다음과 같이 구합니다.
    (당일 종가 + 1일 전 종가 + 2일 전 종가 + 3일 전 종가 + 4일 전 종가) / 5

    :param price: (가격데이터) 단순 이동평균을 구하는데 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 단순 이동평균을 구하는데 사용하는 기간
    :return:
    """
    return price.rolling(window=period).mean()


def ema(price: Series, period: int) -> Series:
    """
    지수이동평균

    <설명>
    지수 이동평균을 구하는 함수입니다.
    지수 이동평균은 가격(price)의 최근 데이터에 가중치를 두어 평균을 구합니다.

    <사용 법>
    첫 번째 인자에는 지수 이동평균을 구하는데 사용하는 가격을,
    두 번째 인자에는 지수 이동평균을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 5일간 종가의 지수 이동평균을 구하고자 하는 경우
    'ema(close, 5)' 또는 '지수이동평균(종가, 5)'와 같이 작성하면 됩니다.

    <계산 방법>
    5일간 종가의 지수 이동평균은 다음과 같이 구합니다.
    당일 지수 이동평균 = 당일 종가 * 평활 계수 + 전일 지수 이동평균 * (1 - 평활 계수)
    평활 계수 = 2 / (5 + 1)

    :param price: (가격데이터) 지수 이동평균을 구하는데 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 지수 이동평균을 구하는데 사용하는 기간
    :return:
    """
    return price.ewm(span=period, min_periods=period, adjust=False).mean()


def ewma(price: Series, period: int) -> Series:
    return price.ewm(span=period, min_periods=period, adjust=True).mean()


def wma(price: Series, period: int) -> Series:
    """
    가중이동평균

    <설명>
    가중 이동평균을 구하는 함수입니다.
    가중 이동평균은 평균을 구하는 데 있어서 주어지는 가중값을 반영시킵니다.

    <사용 방법>
    첫 번째 인자에는 가중 이동평균을 구하는데 사용하는 가격을,
    두 번째 인자에는 가중 이동평균을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 5일간 종가의 가중 이동평균을 구하고자 하는 경우
    'wma(close, 5)' 또는 '가중이동평균(종가, 5)'와 같이 작성하면 됩니다.

    <계산 방법>
    5일간 종가의 가중 이동평균은 다음과 같이 구합니다.
    (당일 종가 * 5 + 1일 전 종가 * 4 + 2일 전 종가 * 3 + 3일 전 종가 * 2 + 4일 전 종가 * 1) / (5 + 4 + 3 + 2 + 1)

    :param price: (가격데이터) 가중 이동평균을 구하는데 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 가중 이동평균을 구하는데 사용하는 기간
    :return:
    """
    weight = np.arange(1, period + 1)
    return price.rolling(window=period).apply(lambda prices: np.dot(prices, weight) / weight.sum(), raw=True)


def _bollinger_band(
        price: Series, period: int, moving_average: MovingAverage, multiplier: int
) -> Tuple[Series, Series]:
    if moving_average == MovingAverage.sma:
        line_mid = sma(price, period)
    elif moving_average == MovingAverage.ema:
        line_mid = ema(price, period)
    elif moving_average == MovingAverage.ewma:
        line_mid = ewma(price, period)
    elif moving_average == MovingAverage.wma:
        line_mid = wma(price, period)
    line_std = price.rolling(window=period).std(ddof=0)
    bollinger_range = line_std.multiply(multiplier)

    return line_mid, bollinger_range


def bollinger_upper(price: Series, period: int, moving_average: MovingAverage, multiplier: int) -> Series:
    """
    볼린저밴드상한선

    <설명>
    볼린저 밴드 상한선(상향 밴드)을 구하는 함수입니다.
    볼린저 밴드의 상한선은 표준 편차에 의해 산출된 이동평균 값이며,
    주가나 지수의 움직임이 큰 시기에는 밴드의 폭이 넓어지고 움직임이 작은 시기에는 밴드의 폭이 좁아지는 특성을 가지고 있습니다.
    즉, 가격 움직임의 크기에 따라 밴드의 넓이가 결정되는 것입니다.

    <사용 방법>
    첫 번째 인자에는 상향 밴드를 구하는데 사용하는 가격을,
    두 번째 인자에는 상향 밴드를 구하는데 사용하는 기간을,
    세 번째 인자에는 중간 밴드를 구하는데 사용하는 이동평균 종류를,
    네 번째 인자에는 상향 밴드를 구할 때 사용하는 표준편차 승수를 적으면 됩니다.
    예를 들어, 20일간 종가의 단순 이동평균으로 중간 밴드를 구하고 상향 밴드는 중간 밴드에 20일간 종가의 표준편차에 2배를 더한 값을 사용하고자 할 경우
    'bollinger_upper(close, 20, sma, 2)' 또는 '볼린저밴드상한선(종가, 20, 단순이동평균, 2)'와 같이 작성하면 됩니다.

    :param price: (가격데이터) 상향 밴드를 구하는데 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 상향 밴드를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 중간 밴드를 구할 때 사용하는 이동평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :param multiplier: (승수값) 상향 밴드를 구할 때 사용하는 표준편차 승수
    :return:
    """
    line_mid, bollinger_range = _bollinger_band(price, period, moving_average, multiplier)
    return line_mid + bollinger_range


def bollinger_lower(price: Series, period: int, moving_average: MovingAverage, multiplier: int) -> Series:
    """
    볼린저밴드하한선

    <설명>
    볼린저 밴드 하한선(하향 밴드)을 구하는 함수입니다.
    볼린저 밴드의 하한선은 표준 편차에 의해 산출된 이동평균 값이며,
    주가나 지수의 움직임이 큰 시기에는 밴드의 폭이 넓어지고 움직임이 작은 시기에는 밴드의 폭이 좁아지는 특성을 가지고 있습니다.
    즉, 가격 움직임의 크기에 따라 밴드의 넓이가 결정되는 것입니다.

    <사용 방법>
    첫 번째 인자에는 하향 밴드를 구하는데 사용하는 가격을,
    두 번째 인자에는 하향 밴드를 구하는데 사용하는 기간을,
    세 번째 인자에는 중간 밴드를 구하는데 사용하는 이동평균 종류를,
    네 번째 인자에는 하향 밴드를 구할 때 사용하는 표준편차 승수를 적으면 됩니다.
    예를 들어, 20일간 종가의 단순 이동평균으로 중간 밴드를 구하고 하향 밴드는 중간 밴드에 20일간 종가의 표준편차에 2배를 뺀 값을 사용하고자 할 경우
    'bollinger_lower(close, 20, sma, 2)' 또는 '볼린저밴드하한선(종가, 20, 단순이동평균, 2)'와 같이 작성하면 됩니다.

    :param price: (가격데이터) 하향 밴드를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 하향 밴드를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 중간 밴드를 구할 때 사용하는 이동평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :param multiplier: (승수값) 하향 밴드를 구할 때 사용하는 표준편차 승수
    :return:
    """
    line_mid, bollinger_range = _bollinger_band(price, period, moving_average, multiplier)
    return line_mid - bollinger_range


def _envelope(price: Series, period: int, moving_average: MovingAverage, ratio: float) -> Tuple[Series, Series]:
    if moving_average == MovingAverage.sma:
        line_mid = sma(price, period)
    elif moving_average == MovingAverage.ema:
        line_mid = ema(price, period)
    elif moving_average == MovingAverage.ewma:
        line_mid = ewma(price, period)
    elif moving_average == MovingAverage.wma:
        line_mid = wma(price, period)

    envelope_range = line_mid.multiply(ratio)
    return line_mid, envelope_range


def envelope_upper(price: Series, period: int, moving_average: MovingAverage, ratio: float) -> Series:
    """
    엔벨로프상한선

    <설명>
    엔벨로프 상한선을 구하는 함수입니다.
    엔벨로프 상한선은 주가의 이동평균 선에서 일정 비율 만큼 더한 선입니다.

    <사용 방법>
    첫 번째 인자에는 엔벨로프 상한선을 구하는데 사용하는 가격을,
    두 번째 인자에는 엔벨로프 상한선을 구하는데 사용하는 기간을,
    세 번째 인자에는 엔벨로프 상한선을 구하는데 사용하는 이동평균 종류를,
    네 번째 인자에는 엔벨로프 상한선을 구할 때 사용하는 비율을 적으면 됩니다.
    예를 들어, 20일간 종가의 단순 이동평균으로 기준선을 구하고 엔벨로프 상한선은 기준선에서 8% 위의 선으로 하고자 하는 경우
    'envelope_upper(close, 20, sma, 0.08)' 또는 엔벨로프상한선(종가, 20, 단순이동평균, 0.08)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 엔벨로프 상한선을 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 엔벨로프 상한선을 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 엔벨로프 상한선을 구하는데 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :param ratio: (비율) 엔벨로프 상한선을 구하는데 사용하는 비율
    :return:
    """
    line_mid, envelope_range = _envelope(price, period, moving_average, ratio)
    return line_mid + envelope_range


def envelope_lower(price: Series, period: int, moving_average: MovingAverage, ratio: float) -> Series:
    """
    엔벨로프하한선

    <설명>
    엔벨로프 하한선을 구하는 함수입니다.
    엔벨로프 하한선은 주가의 이동평균 선에서 일정 비율 만큼 뺀 선입니다.

    <사용 방법>
    첫 번째 인자에는 엔벨로프 하한선을 구하는데 사용하는 가격을,
    두 번째 인자에는 엔벨로프 하한선을 구하는데 사용하는 기간을,
    세 번째 인자에는 엔벨로프 하한선을 구하는데 사용하는 이동평균 종류를,
    네 번째 인자에는 엔벨로프 하한선을 구할 때 사용하는 비율을 적으면 됩니다.
    예를 들어, 20일간 종가의 단순 이동평균으로 기준선을 구하고 엔벨로프 하한선은 기준선에서 8% 아래의 선으로 하고자 하는 경우
    'envelope_lower(close, 20, sma, 0.08)' 또는 '엔벨로프하한선(종가, 20, 단순이동평균, 0.08)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 엔벨로프 하한선을 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 엔벨로프 하한선을 구하는 기간
    :param moving_average: (이동평균종류) 엔벨로프 하한선을 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :param ratio: (비율) 엔벨로프 상한선을 구하는데 사용하는 비율
    :return:
    """
    line_mid, envelope_range = _envelope(price, period, moving_average, ratio)
    return line_mid - envelope_range


def pivot_standard(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇기준선

    <설명>
    피봇기준선을 구하는 함수입니다.
    피봇기준선은 전일 고가, 저가, 종가의 평균입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇기준선을 구하고자 하는 경우
    'pivot_standard(high, low, close)' 또는 '피봇기준선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    return (price_high + price_low + price_close).shift(1) / 3


def pivot_standard_for_sellprice(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇기준선

    <설명>
    피봇기준선을 구하는 함수입니다.
    피봇기준선은 전일 고가, 저가, 종가의 평균입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇기준선을 구하고자 하는 경우
    'pivot_standard(high, low, close)' 또는 '피봇기준선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    return (price_high + price_low + price_close) / 3


def pivot_second_upper(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇2차저항선

    <설명>
    피봇2차저항선을 구하는 함수입니다.
    피봇2차저항선은 피봇 기준선에 전일 변동폭을 더한 값입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇2차저항선을 구하고자 하는 경우
    'pivot_second_upper(high, low, close)' 또는 '피봇2차저항선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    pivot = pivot_standard(price_high, price_low, price_close)
    return pivot + price_high.shift(1).sub(price_low.shift(1))


def pivot_first_upper(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇1차저항선

    <설명>
    피봇1차저항선을 구하는 함수입니다.
    피봇1차저항선은 피봇 기준선에 두배를 곱한 후 전일 저가를 뺀 값입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇1차저항선을 구하고자 하는 경우
    'pivot_first_upper(high, low, close)' 또는 '피봇1차저항선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    pivot = pivot_standard(price_high, price_low, price_close)
    return (pivot * 2).sub(price_low.shift(1))


def pivot_first_lower(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇1차지지선

    <설명>
    피봇1차지지선을 구하는 함수입니다.
    피봇1차지지선은 피봇 기준선에 두배를 곱한 후 전일 고가를 뺀 값입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇1차지지선을 구하고자 하는 경우
    'pivot_first_lower(high, low, close)' 또는 '피봇1차지지선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    pivot = pivot_standard(price_high, price_low, price_close)
    return (pivot * 2).sub(price_high.shift(1))


def pivot_second_lower(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    피봇2차지지선

    <설명>
    피봇2차지지선을 구하는 함수입니다.
    피봇2차지지선은 피봇 기준선에 전일 변동폭을 뺀 값입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    피봇2차지지선 구하고자 하는 경우
    'pivot_second_lower(high, low, close)' 또는 '피봇2차지지선(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    pivot = pivot_standard(price_high, price_low, price_close)
    return pivot.sub(price_high.shift(1)) + price_low.shift(1)


def price_channel_upper(price_high: Series, period: int) -> Series:
    """
    가격채널상한선

    <설명>
    가격 채널 상한선을 구하는 함수입니다.
    가격 채널 상한선은 일정 기간 내의 최고가를 이은 선입니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 가격 채널 상한선을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 20일간 채널 지표 상한선을 구하고자 하는 경우
    'price_channel_upper(high, 20)' 또는 '가격채널상한선(고가, 20)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param period: (기간) 가격 채널 상한선을 구할 때 사용하는 기간
    :return:
    """
    return price_high.shift(1).rolling(window=period).max()


def price_channel_lower(price_low: Series, period: int) -> Series:
    """
    가격채널하한선

    <설명>
    가격 채널 하한선을 구하는 함수입니다.
    가격 채널 하한선은 일정 기간 내의 최저가를 이은 선입니다.

    <사용 방법>
    첫 번째 인자에는 저가를,
    두 번째 인자에는 가격 채널 하한선을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 20일간 채널 지표 하한선을 구하고자 하는 경우
    'price_channel_lower(low, 20)' 또는 '가격채널하한선(저가, 20)'과 같이 작성하면 됩니다.

    :param price_low: (저가) 저가
    :param period: (기간) 가격 채널 상한선을 구하는 기간
    :return:
    """
    return price_low.shift(1).rolling(window=period).min()


def _tr(price_high: Series, price_low: Series, price_close: Series) -> Series:
    data = {
        "range": price_high - price_low,
        "up": abs(price_high - price_close.shift(1)),
        "down": abs(price_close.shift(1) - price_low),
    }
    data = DataFrame(data, columns=["range", "up", "down"])

    return data.max(axis=1)


def pdi(
        price_high: Series, price_low: Series, price_close: Series, period: int, moving_average: MovingAverage
) -> Series:
    """
    매수방향지표

    <설명>
    매수방향지표(PDI)를 구하는 함수입니다.
    매수방향지표(PDI)는 실질적으로 상승하는 폭의 비율을 나타냅니다.
    매수방향지표(PDI)는 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 매수방향지표(PDI)를 구하는데 사용하는 기간을,
    다섯 번째 인자에는 매수방향지표(PDI)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 지수 이동 평균을 이용한 14일간 매수방향지표(PDI)를 구하고자 하는 경우
    'pdi(high, low, close, 14, ema)' 또는 '매수방향지표(고가, 저가, 종가, 14, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) 매수방향지표(PDI)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 매수방향지표(PDI)를 구하는데 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    pdm = np.where(
        ((price_high.diff(1) > 0) & (price_high.diff(1) > price_low.shift(1) - price_low)), price_high.diff(1), 0
    )

    if moving_average == MovingAverage.sma:
        pdmn = sma(Series(pdm), period)
    elif moving_average == MovingAverage.ema:
        pdmn = ema(Series(pdm), period)
    elif moving_average == MovingAverage.ewma:
        pdmn = ewma(Series(pdm), period)
    elif moving_average == MovingAverage.wma:
        pdmn = wma(Series(pdm), period)

    tr = _tr(price_high, price_low, price_close)

    if moving_average == MovingAverage.sma:
        trn = sma(tr, period)
    elif moving_average == MovingAverage.ema:
        trn = ema(tr, period)
    elif moving_average == MovingAverage.ewma:
        trn = ewma(tr, period)
    elif moving_average == MovingAverage.wma:
        trn = wma(tr, period)

    return pdmn.divide(trn)


def mdi(
        price_high: Series, price_low: Series, price_close: Series, period: int, moving_average: MovingAverage
) -> Series:
    """
    매도방향지표

    <설명>
    매도방향지표(MDI)를 구하는 함수입니다.
    매도방향지표(MDI)는 실질적으로 하락하는 폭의 비율을 나타냅니다.
    매도방향지표(MDI)는 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 매도방향지표(MDI)를 구하는데 사용하는 기간을,
    다섯 번째 인자에는 매도방향지표(MDI)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 지수 이동 평균을 이용한 14일간 매도방향지표(MDI)를 구하고자 하는 경우
    'mdi(high, low, close, 14, ema)' 또는 '매도방향지표(고가, 저가, 종가, 14, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) 매도방향지표(MDI)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 매도방향지표(MDI)를 구하는데 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    mdm = np.where(
        (((price_low.shift(1) - price_low) > 0) & (price_high.diff(1) < (price_low.shift(1) - price_low))),
        price_low.shift(1) - price_low,
        0,
    )

    if moving_average == MovingAverage.sma:
        mdmn = sma(Series(mdm), period)
    elif moving_average == MovingAverage.ema:
        mdmn = ema(Series(mdm), period)
    elif moving_average == MovingAverage.ewma:
        mdmn = ewma(Series(mdm), period)
    elif moving_average == MovingAverage.wma:
        mdmn = wma(Series(mdm), period)

    tr = _tr(price_high, price_low, price_close)

    if moving_average == MovingAverage.sma:
        trn = sma(tr, period)
    elif moving_average == MovingAverage.ema:
        trn = ema(tr, period)
    elif moving_average == MovingAverage.ewma:
        trn = ewma(tr, period)
    elif moving_average == MovingAverage.wma:
        trn = wma(tr, period)

    return mdmn.divide(trn)


def adx(
        price_high: Series, price_low: Series, price_close: Series, period: int, moving_average: MovingAverage
) -> Series:
    """
    평균방향이동지표

    <설명>
    평균방향이동지표(ADX)를 구하는 함수입니다.
    평균방향이동지표(ADX)는 추세의 강도를 의미합니다.
    평균방향이동지표(ADX)는 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 ADX를 구하는데 사용하는 기간을,
    다섯 번째 인자에는 ADX를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 지수 이동 평균을 이용한 14일간 ADX를 구하고자 하는 경우
    'adx(high, low, close, 14, ema)' 또는 '평균방향이동지표(고가, 저가, 종가, 14, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) ADX를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) ADX를 구하는데 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """

    pdi_val = pdi(price_high, price_low, price_close, period, moving_average)
    mdi_val = mdi(price_high, price_low, price_close, period, moving_average)
    dx = abs(pdi_val - mdi_val).divide(pdi_val + mdi_val)

    if moving_average == MovingAverage.sma:
        return sma(dx, period)
    elif moving_average == MovingAverage.ema:
        return ema(dx, period)
    elif moving_average == MovingAverage.ewma:
        return ewma(dx, period)
    elif moving_average == MovingAverage.wma:
        return wma(dx, period)


def macd(price: Series, short_period: int, long_period: int, moving_average: MovingAverage) -> Series:
    """
    이동평균수렴확산지수

    <설명>
    이동평균수렴확산지수(MACD)를 구하는 함수입니다.
    이동평균수렴확산지수(MACD)는 단기 이동평균 값과 장기 이동평균 값의 차이를 이용한 지표입니다.

    <사용 방법>
    첫 번째 인자에는 이동평균수렴확산지수(MACD)를 구하는데 사용하는 가격을,
    두 번째 인자에는 단기 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 장기 이동 평균을 구하는데 사용하는 기간을,
    네 번째 인자에는 이동평균수렴확산지수(MACD)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 12일간 종가의 단순 이동 평균과 26일간 종가의 단순 이동 평균을 이용하여 이동평균수렴확산지수(MACD)를 구하고자 하는 경우에는
    'macd(close, 12, 26, sma)' 또는 '이동평균수렴확산지수(종가, 12, 26, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 이동평균수렴확산지수(MACD)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param short_period: (단기이동평균기간) 단기 이동 평균을 구하는데 사용하는 기간
    :param long_period: (장기이동평균기간) 장기 이동 평균을 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 이동평균수렴확산지수(MACD)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    if moving_average == MovingAverage.sma:
        short_term = sma(price, short_period)
        long_term = sma(price, long_period)
    elif moving_average == MovingAverage.ema:
        short_term = ema(price, short_period)
        long_term = ema(price, long_period)
    elif moving_average == MovingAverage.ewma:
        short_term = ewma(price, short_period)
        long_term = ewma(price, long_period)
    elif moving_average == MovingAverage.wma:
        short_term = wma(price, short_period)
        long_term = wma(price, long_period)

    return short_term - long_term


def macd_signal(
        price: Series, short_period: int, long_period: int, signal_period: int, moving_average: MovingAverage
) -> Series:
    """
    이동평균수렴확산시그널

    <설명>
    이동평균수렴확산시그널(MACD_signal)을 구하는 함수입니다.
    이동평균수렴확산시그널(MACD_signal)은 이동평균수렴확산지수(MACD)의 일정 기간 동안의 평균입니다.

    <사용 방법>
    첫 번째 인자에는 이동평균수렴확산시그널(MACD_signal)을 구하는데 사용하는 가격을,
    두 번째 인자에는 단기 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 장기 이동 평균을 구하는데 사용하는 기간을,
    네 번째 인자에는 시그널 기간을,
    다섯 번째 인자에는 이동평균수렴확산시그널(MACD_signal)을 구하는데 이용할 이동 평균 종류를 적으면 됩니다.
    예를 들어, 12일간 종가의 단순 이동 평균, 26일간 종가의 단순 이동 평균, 9일의 signal 기간을 이용하여
    이동평균수렴확산시그널(MACD_signal)을 구하고자 하는 경우에는
    'macd_signal(close, 12, 26, 9, sma)' 또는 '이동평균수렴확산시그널(종가, 12, 26, 9, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price: (기간) 이동평균수렴확산시그널(MACD_signal)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param short_period: (단기이동평균기간) 단기 이동 평균을 구하는데 사용하는 기간
    :param long_period: (장기이동평균기간) 장기 이동 평균을 구하는데 사용하는 기간
    :param signal_period: (시그널기간) 이동평균수렴확산시그널(MACD_signal)를 구할 때 사용하는 시그널 기간
    :param moving_average: (이동평균종류) 이동평균수렴확산시그널(MACD_signal)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    macd_val = macd(price, short_period, long_period, moving_average)
    if moving_average == MovingAverage.sma:
        signal = sma(macd_val, signal_period)
    elif moving_average == MovingAverage.ema:
        signal = ema(macd_val, signal_period)
    elif moving_average == MovingAverage.ewma:
        signal = ewma(macd_val, signal_period)
    elif moving_average == MovingAverage.wma:
        signal = wma(macd_val, signal_period)

    return signal


def macd_oscillator(
        price: Series, short_period: int, long_period: int, signal_period: int, moving_average: MovingAverage
) -> Series:
    """
    이동평균수렴확산오실레이터

    <설명>
    이동평균수렴확산오실레이터(MACD_oscillator)를 구하는 함수입니다.
    이동평균수렴확산오실레이터(MACD_oscillator)는 MACD와 Signal의 차를 통해 계산됩니다.

    <사용 방법>
    첫 번째 인자에는 이동평균수렴확산오실레이터(MACD_oscillator)를 구하는데 사용하는 가격을,
    두 번째 인자에는 단기 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 장기 이동 평균을 구하는데 사용하는 기간을,
    네 번째 인자에는 시그널 기간을,
    다섯 번째 인자에는 이동평균수렴확산오실레이터(MACD_oscillator)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 12일간 종가의 단순 이동 평균, 26일간 종가의 단순 이동 평균, 9일의 signal 기간을 이용하여
    이동평균수렴확산오실레이터(MACD_oscillator)를 구하고자 하는 경우에는
    'macd_oscillator(close, 12, 26, 9, sma)' 또는 '이동평균수렴확산오실레이터(종가, 12, 26, 9, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 이동평균수렴확산오실레이터(MACD_oscillator)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param short_period: (단기이동평균기간) 단기 이동 평균을 구하는데 사용하는 기간
    :param long_period: (장기이동평균기간) 장기 이동 평균을 구하는데 사용하는 기간
    :param signal_period: (시그널기간) 이동평균수렴확산오실레이터(MACD_oscillator)를 구할 때 사용하는 시그널 기간
    :param moving_average: (이동평균종류) 이동평균수렴확산오실레이터(MACD_oscillator)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    macd_val = macd(price, short_period, long_period, moving_average)
    if moving_average == MovingAverage.sma:
        signal = sma(macd_val, signal_period)
    elif moving_average == MovingAverage.ema:
        signal = ema(macd_val, signal_period)
    elif moving_average == MovingAverage.ewma:
        signal = ewma(macd_val, signal_period)
    elif moving_average == MovingAverage.wma:
        signal = wma(macd_val, signal_period)

    return macd_val - signal


def stochastic_fast_k(price_high: Series, price_low: Series, price_close: Series, k_period: int) -> Series:
    """
    스토캐스틱

    <설명>
    스토캐스틱(stochastic_fast_k)을 구하는 함수입니다.
    정해진 기간동안(k_period)의 가격 범위에서 오늘의 시장가격이 상대적으로 어디에 위치하고 있는지를 알려주는 지표로써,
    시장가격이 상승추세에 있다면 현재가격은 최고가 부근에 위치할 가능성이 높고,
    하락추세에 있다면 현재가는 최저가 부근에서 형성될 가능성이 높다는 것에 착안하여 만들어진 지표입니다.

    <사용방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 스토캐스틱(stochastic_fast_k)을 구하는데 사용할 기간을 적으면 됩니다.
    예를 들어, 20일간의 스토캐스틱(stochastic_fast_k)을 구하고자 하는 경우에는
    'stochastic_fast_k(high, low, close, 20)' 또는 '스토캐스틱(고가, 저가, 종가, 20)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param k_period: (스토캐스틱기간) 스토캐스틱(stochastic_fast_k)을 구하는데 사용할 기간
    :return:
    """
    return (price_close - price_low.rolling(window=k_period).min()) / (
            price_high.rolling(window=k_period).max() - price_low.rolling(window=k_period).min()
    )


def stochastic_fast_d(
        price_high: Series,
        price_low: Series,
        price_close: Series,
        k_period: int,
        d_period: int,
        moving_average: MovingAverage,
) -> Series:
    """
    스토캐스틱이동평균

    <설명>
    스토캐스틱이동평균(stochastic_fast_d)을 구하는 함수입니다.
    스토캐스틱이동평균은 스토캐스틱을 정해진 기간(d_period)동안 평균을 낸 값입니다.

    <사용방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 스토캐스틱(stochastic_fast_k)을 구하는데 사용할 기간을,
    다섯 번째 인자에는 스토캐스틱이동평균(stochastic_fast_d)을 구하는데 사용한 기간을,
    여섯 번째 인자에는 스토캐스틱이동평균(stochastic_fast_d)을 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 20일간의 스토캐스틱(stochastic_fast_k)을 구하고 이를 5일간 단순 이동평균을 낸 값을 사용하고자 하는 경우에는
    'stochastic_fast_d(high, low, close, 20, 5, sma)' 또는 '스토캐스틱이동평균(고가, 저가, 종가, 20, 5, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param k_period: (스토캐스틱기간) 스토캐스틱(stochastic_fast_k)을 구하는데 사용할 기간
    :param d_period: (이동평균기간) 스토캐스틱이동평균(stochastic_fast_d)을 구하는데 사용할 기간
    :param moving_average (이동평균종류) 스토캐스틱이동평균(stochastic_fast_d)을 구하는데 사용할 이동평균 종류
    :return:
    """
    stochastic_fast_k_val = stochastic_fast_k(price_high, price_low, price_close, k_period)

    if moving_average == MovingAverage.sma:
        return sma(stochastic_fast_k_val, d_period)
    elif moving_average == MovingAverage.ema:
        return ema(stochastic_fast_k_val, d_period)
    elif moving_average == MovingAverage.ewma:
        return ewma(stochastic_fast_k_val, d_period)
    elif moving_average == MovingAverage.wma:
        return wma(stochastic_fast_k_val, d_period)


def volume_ratio(price: Series, volume: Series, period: int) -> Series:
    """
    거래량비율

    <설명>
    거래량비율(Volume Ratio)을 구하는 함수입니다.
    거래량비율(Volume Ratio)은 일정 기간 동안의 상승일의 거래량과 하락일의 거래량을 비교합니다.
    거래량비율(Volume Ratio)은 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 거래량비율(Volume Ratio)을 구하는데 사용하는 가격을,
    두 번째 인자에는 거래량을,
    세 번째 인자에는 거래량비율(Volume Ratio)을 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 20일간의 종가를 이용한 거래량비율(Volume Ratio)을 구하고자 하는 경우에는
    'volume_ratio(close, volume, 20)' 또는 '거래량비율(종가, 거래량, 20)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 거래량비율(Volume Ratio)을 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param volume: (거래량) 거래량
    :param period: (기간) 거래량비율(Volume Ratio)을 구하는데 사용하는 기간
    :return:
    """
    up = np.where(price.diff(1).gt(0), volume, 0)
    down = np.where(price.diff(1).lt(0), volume, 0)
    maintain = np.where(price.diff(1).equals(0), volume.mul(0.5), 0)

    up = up + maintain
    down = down + maintain
    sum_up = Series(up).rolling(window=period, min_periods=period).sum()
    sum_down = Series(down).rolling(window=period, min_periods=period).sum()
    return sum_up.div(sum_down)


def 순거래대금(price: Series, volume: Series, period: int) -> Series:

    up = np.where(price.diff(1).gt(0), price * volume, 0)
    down = np.where(price.diff(1).lt(0), -1 * price * volume, 0)
    sum = up + down

    rolling_sum = Series(sum).rolling(window=period, min_periods=period).sum()
    return rolling_sum


def psychological_line(price: Series, period: int) -> Series:
    """
    투자심리도

    <설명>
    투자심리도(Psychological Line)를 구하는 함수입니다.
    투자심리도(Psychological Line)를 이용하면 과열 및 침체도를 파악할 수 있습니다.
    투자심리도(Psychological Line)는 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 투자심리도(Psychological Line)를 구하는데 사용하는 가격을,
    두 번째 인자에는 투자심리도(Psychological Line)를 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 10일간의 종가를 이용한 투자심리도(Psychological Line)를 구하고자 하는 경우에는
    'psychological_line(close, 10)' 또는 '투자심리도(종가, 10)'과 같이 작성하면 됩니다.

    <계산 방법>
    10일간의 종가를 이용한 투자심리도(Psychological Line)는 다음과 같이 구합니다.
    (10일간 전일 종가 대비 상승 일수) / 10

    :param price: (가격데이터) 투자심리도(Psychological Line)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 투자심리도(Psychological Line)를 구하는데 사용하는 기간
    :return:
    """
    up = np.where(price.diff(1).gt(0), 1, 0)
    sum_up = Series(up).rolling(window=period, min_periods=period).sum()

    return sum_up.divide(period)


def new_psychological_line(price: Series, period: int) -> Series:
    """
    신심리도

    <설명>
    신심리도(Psychological Line)를 구하는 함수입니다.
    신심리도(New Psychological Line)는 주가 등락 폭을 반영하지 못하는 투자심리도(Psychological Line)의 단점을 개선하였습니다.

    <사용 방법>
    첫 번째 인자에는 신심리도(New Psychological Line)를 구하는데 사용하는 가격을,
    두 번째 인자에는 신심리도(New Psychological Line)를 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 10일간의 종가를 이용한 신심리도(New Psychological Line)를 구하고자 하는 경우에는
    'new_psychological_line(close, 10)' 또는 '신심리도(종가, 10)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 신심리도(New Psychological Line)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 신심리도(New Psychological Line)를 구하는데 사용하는 기간
    :return:
    """
    up_cnt = np.where(price.diff(1).gt(0), 1, 0)
    up_cnt_cum = Series(up_cnt).rolling(window=period, min_periods=period).sum()
    up_width = np.where(price.diff(1).gt(0), price.diff(1), 0)
    up_width_cum = Series(up_width).rolling(window=period, min_periods=period).sum()

    down_cnt = np.where(price.diff(1).lt(0), 1, 0)
    down_cnt_cum = Series(down_cnt).rolling(window=period, min_periods=period).sum()
    down_width = np.where(price.diff(1).lt(1), abs(price.diff(1)), 0)
    down_width_cum = Series(down_width).rolling(window=period, min_periods=period).sum()

    up = up_cnt_cum.multiply(up_width_cum)
    down = down_cnt_cum.multiply(down_width_cum)

    quo = up.subtract(down)
    deno = (up_width_cum + down_width_cum).multiply(period)

    return quo.divide(deno)


def disparity(price: Series, period: int, moving_average: MovingAverage) -> Series:
    """
    이격도

    <설명>
    이격도(Disparity)를 구하는 함수입니다.
    이격도(Disparity)는 주가가 이동 평균과 어느 정도 떨어져 있는가 나타냅니다.

    <사용 방법>
    첫 번째 인자에는 이격도(Disparity)를 구하는데 사용하는 가격을,
    두 번째 인자에는 이격도(Disparity)를 구하는데 사용하는 기간을,
    세 번째 인자에는 이격도(Disparity)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 종가와 5일간의 단순 이동평균을 이용한 이격도(Disparity)를 구하고자 하는 경우에는
    'disparity(close, 5, sma)' 또는 '이격도(종가, 5, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 이격도(Disparity)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 이격도(Disparity)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 이격도(Disparity)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    if moving_average == MovingAverage.sma:
        ma = sma(price, period)
    elif moving_average == MovingAverage.ema:
        ma = ema(price, period)
    elif moving_average == MovingAverage.wma:
        ma = wma(price, period)
    return price.divide(ma)


def ibs(price_high: Series, price_low: Series, price_close: Series) -> Series:
    """
    종가위치비율

    <설명>
    종가위치비율(IBS)을 구하는 함수입니다.
    종가위치비율(IBS)은 종가가 당일 변동폭에서 어떠한 지점에 위치해있는지를 나타냅니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가,
    세 번째 인자에는 종가를 적으면 됩니다.
    종가위치비율(IBS)을 구하고자 하는 경우에는
    'ibs(high, low, close)' 또는 '종가위치비율(고가, 저가, 종가)'와 같이 작성하면 됩니다.

    <계산 방법>
    (종가 - 저가) / (고가 - 저가)

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    return (price_close - price_low) / (price_high - price_low)


def upper_tail_ratio(price_open: Series, price_high: Series, price_close: Series) -> Series:
    """
    윗꼬리비율

    <설명>
    윗꼬리비율(Upper Tail Ratio)을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자에는 시가를,
    두 번째 인자에는 고가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    윗꼬리비율은 'upper_tail_ratio(open, high, close)' 또는 '윗꼬리비율(시가, 고가, 종가)'과 같이 작성하면 됩니다.

    :param price_open: (시가) 시가
    :param price_high: (고가) 고가
    :param price_close: (종가) 종가
    :return:
    """
    price_upper = np.where(price_open >= price_open, price_open, price_close)
    price_upper = Series(price_upper)

    return (price_high - price_upper) / (price_close - price_open).abs()


def lower_tail_ratio(price_open: Series, price_low: Series, price_close: Series) -> Series:
    """
    아랫꼬리비율

    <설명>
    아랫꼬리비율(Lower Tail Ratio)을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자에는 시가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를 적으면 됩니다.
    아랫꼬리비율을 구하고자 하는 경우에는
    'lower_tail_ratio(open, low, close)' 또는 '아랫꼬리비율(시가, 저가, 종가)'과 같이 작성하면 됩니다.

    :param price_open: (시가) 시가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :return:
    """
    price_under = np.where(price_open >= price_open, price_close, price_open)
    price_under = Series(price_under)

    return (price_under - price_low) / (price_close - price_open).abs()


def a_ratio(price_open: Series, price_high: Series, price_low: Series, period: int) -> Series:
    """
    에이비율

    <설명>
    A Ratio를 구하는 함수입니다.
    A Ratio는 주가 변동을 이용하여 강, 약 에너지를 파악합니다.

    <사용 방법>
    첫 번째 인자에는 시가를,
    두 번째 인자에는 고가를,
    세 번째 인자에는 저가를,
    네 번째 인자에는 A Ratio를 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 26일간의 A Ratio를 구하고자 하는 경우에는
    'a_ratio(open, high, low, 26)' 또는 '에이비율(시가, 고자, 저가, 26)'과 같이 작성하면 됩니다.

    :param price_open: (시가) 시가
    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param period: (기간) A Ratio를 구하는데 사용하는 기간
    :return:
    """
    strength = price_high - price_open
    weakness = price_open - price_low
    return strength.rolling(window=period).sum() / weakness.rolling(window=period).sum()


def b_ratio(price_high: Series, price_low: Series, price_close: Series, period: int) -> Series:
    """
    비비율

    <설명>
    B Ratio를 구하는 함수입니다.
    B Ratio는 주가 변동을 이용하여 강, 약 에너지를 파악합니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 B Ratio를 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 26일간의 B Ratio를 구하고자 하는 경우에는
    'b_ratio(high, low, close, 26)' 또는 '비비율(고가, 저가, 종가, 26)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) B Ratio를 구하는데 사용하는 기간
    :return:
    """
    strength = price_high - price_close.shift(1)
    weakness = price_close.shift(1) - price_low
    return strength.rolling(window=period).sum() / weakness.rolling(window=period).sum()


def mass_index(
        price_high: Series, price_low: Series, period: int, moving_average: MovingAverage = MovingAverage.ewma
) -> Series:
    """
    질량지수

    <설명>
    질량지수(Mass Index)를 구하는 함수입니다.
    질량지수(Mass Index)는 고가와 저가 사이의 변동폭을 측정하여 단기적인 추세의 전환점을 찾아냅니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 질량지수(Mass Index)를 구하는데 사용하는 기간을,
    네 번째 인자에는 질량지수(Mass Index)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 단순 이동평균을 이용한 25일간 질량지수(Mass Index)를 구하고자 하는 경우에는
    'mass_index(high, low, 25, sma)' 또는 '질량지수(고가, 저가, 25, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param period: (기간) 질량지수(Mass Index)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 질량지수(Mass Index)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    day_range = price_high - price_low

    if moving_average == MovingAverage.sma:
        single = sma(day_range, 9)
        double = sma(single, 9)
    elif moving_average == MovingAverage.ema:
        single = ema(day_range, 9)
        double = ema(single, 9)
    elif moving_average == MovingAverage.ewma:
        single = ewma(day_range, 9)
        double = ewma(single, 9)
    elif moving_average == MovingAverage.wma:
        single = wma(day_range, 9)
        double = wma(single, 9)
    ratio = single / double

    return ratio.rolling(window=period).sum()


def mao(price: Series, short_period: int, long_period: int, moving_average: MovingAverage) -> Series:
    """
    이동평균오실레이터

    <설명>
    이동평균오실레이터(MAO)를 구하는 함수입니다.
    이동평균오실레이터(MAO)는 단기 이동 평균 값과 장기 이동 평균 값의 차를 통해 계산되며 주가의 추세를 판단할 수 있습니다.

    <사용 방법>
    첫 번째 인자에는 이동평균오실레이터(MAO)를 구하는데 사용하는 가격을,
    두 번째 인자에는 단기 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 장기 이동 평균을 구하는데 사용하는 기간을,
    네 번째 인자에는 이동평균오실레이터(MAO)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 5일간 종가의 단순 이동 평균과 10일간 종가의 단순 이동 평균을 사용하여 이동평균오실레이터(MAO)를 구하고자 하는 경우에는
    'mao(close, 5, 10, sma)' 또는 '이동평균오실레이터(종가, 5, 10, 단순이동평균)'과 같이 작성하면 됩니다.

    :param price: (기간) 이동평균오실레이터(MAO)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param short_period: (단기이동평균기간) 단기 이동 평균을 구하는데 사용하는 기간
    :param long_period: (장기이동평균기간) 장기 이동 평균을 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 이동평균오실레이터(MAO)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """

    if moving_average == MovingAverage.sma:
        return sma(price, short_period) - sma(price, long_period)
    elif moving_average == MovingAverage.ema:
        return ema(price, short_period) - ema(price, long_period)
    elif moving_average == MovingAverage.ewma:
        return ewma(price, short_period) - ewma(price, long_period)
    elif moving_average == MovingAverage.wma:
        return wma(price, short_period) - wma(price, long_period)


def sonar(price: Series, period: int, sonar_period: int, moving_average: MovingAverage) -> Series:
    """
    소나

    <설명>
    소나(Sonar)를 구하는 함수입니다.
    소나(Sonar)는 주가의 추세 전환 시점을 파악하기 위한 지표입니다.

    <사용 방법>
    첫 번째 인자에는 소나(Sonar)를 구하는데 사용하는 가격을,
    두 번째 인자에는 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 소나(Sonar)를 구하는데 사용하는 과거 이동 평균 값의 기간을,
    네 번째 인자에는 소나(Sonar)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 20일간 종가의 지수 이동 평균과 9일전 지수 이동 평균을 이용하여 소나(Sonar)를 구하고자 하는 경우에는
    'sonar(close, 20, 9, ema)' 또는 '소나(종가, 20, 9, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 소나(Sonar)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 이동 평균을 구하는데 사용하는 기간
    :param sonar_period: (소나기간) 사용하고자 하는 과거 이동 평균 값의 기간
    :param moving_average: (이동평균종류) 소나(Sonar)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    if moving_average == MovingAverage.sma:
        ma = sma(price, period)
    elif moving_average == MovingAverage.ema:
        ma = ema(price, period)
    elif moving_average == MovingAverage.ewma:
        ma = ewma(price, period)
    elif moving_average == MovingAverage.wma:
        ma = wma(price, period)

    return ma - ma.shift(sonar_period)


def sonar_signal(
        price: Series, period: int, sonar_period: int, signal_period: int, moving_average: MovingAverage
) -> Series:
    """
    소나시그널

    <설명>
    소나시그널(Sonar Signal)을 구하는 함수입니다.
    소나시그널(Sonar Signal)은 주가의 추세 전환 시점을 파악하기 위한 지표입니다.

    <사용 방법>
    첫 번째 인자에는 소나(Sonar)를 구하는데 사용하는 가격을,
    두 번째 인자에는 이동 평균을 구하는데 사용하는 기간을,
    세 번째 인자에는 소나(Sonar)를 구하는데 사용하는 과거 이동 평균 값의 기간을,
    네 번째 인자에는 소나시그널(Sonar Signal)을 구하는데 사용하는 소나(Sonar)의 이동 평균 기간을,
    네 번째 인자에는 소나(Sonar)를 구하는데 이용할 이동 평균 종류를 적으면 됩니다.
    예를 들어, 20일간 종가의 지수 이동 평균과 9일전 지수 이동 평균을 이용하여 시그널 기간이 9인 소나시그널(Sonar Signal)를 구하고자 하는 경우에는
    'sonar_signal(close, 20, 9, 9, ema)' 또는 '소나시그널(종가, 20, 9, 9, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 소나(Sonar)를 구할 때 사용하는 가격 ex) 시가, 고가, 저가, 종가
    :param period: (기간) 이동 평균을 구하는데 사용하는 기간
    :param sonar_period: (소나기간) 사용하고자 하는 과거 이동 평균 값의 기간
    :param signal_period: (시그널기간) 소나시그널(Sonar Signal)을 구하는데 사용하는 소나(Sonar)의 이동 평균 기간
    :param moving_average: (이동평균종류) 소나(Sonar)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    sonar_val = sonar(price, period, sonar_period, moving_average)

    if moving_average == MovingAverage.sma:
        return sma(sonar_val, signal_period)
    elif moving_average == MovingAverage.ema:
        return ema(sonar_val, signal_period)
    elif moving_average == MovingAverage.ewma:
        return ewma(sonar_val, signal_period)
    elif moving_average == MovingAverage.wma:
        return wma(sonar_val, signal_period)


def mfi(price_high: Series, price_low: Series, price_close: Series, vol: Series, period: int) -> Series:
    """
    자금흐름지수

    <설명>
    자금흐름지수(MFI)를 구하는 함수입니다.
    자금흐름지수(MFI)는 주식시장으로 자금이 유입되거나 유출되는 양을 측정합니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 거래량을,
    네 번째 인자에는 자금흐름지수(MFI)를 구하는데 사용하는 기간을 적으면 됩니다.
    예를 들어, 14일간 자금흐름지수(MFI)를 구하고자 하는 경우에는
    'mfi(high, low, close, volume, 14)' 또는 '자금흐름지수(고가, 저가, 종가, 거래량, 14)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param vol: (거래량) 거래량
    :param period: (기간) 자금흐름지수(MFI)를 구하는데 사용하는 기간
    :return:
    """
    typical_price = (price_high + price_low + price_close) / 3
    money_flow = vol * typical_price

    positive_money_flow = np.where(typical_price.diff(1) > 0, money_flow, 0)
    positive_money_flow = Series(positive_money_flow).rolling(window=period, min_periods=period).sum()
    negative_money_flow = np.where(typical_price.diff(1) < 0, money_flow, 0)
    negative_money_flow = Series(negative_money_flow).rolling(window=period, min_periods=period).sum()

    money_flow_ratio = positive_money_flow / negative_money_flow
    return money_flow_ratio / (1 + money_flow_ratio)


def trix(price: Series, period: int, moving_average: MovingAverage) -> Series:
    """
    트라이엄프엑스

    <설명>
    트라이엄프엑스(TRIX)를 구하는 함수입니다.
    트라이엄프엑스(TRIX)는 이동 평균을 세 번 구한 후 이 값의 전일 대비 상승비율을 계산합니다.

    <사용 방법>
    첫 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 가격을,
    두 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 기간을,
    세 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 10일간 종가의 지수 이동 평균으로 트라이엄프엑스(TRIX)를 구하고자 하는 경우에는
    'trix(close, 10, ema)' 또는 '소나시그널(종가, 10, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 트라이엄프엑스(TRIX)를 구하는데 사용하는 가격
    :param period: (기간) 트라이엄프엑스(TRIX)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 트라이엄프엑스(TRIX)를 구할 때 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    if moving_average == MovingAverage.sma:
        ma3 = sma(sma(sma(price, period), period), period)
    elif moving_average == MovingAverage.ema:
        ma3 = ema(ema(ema(price, period), period), period)
    elif moving_average == MovingAverage.ewma:
        ma3 = ewma(ewma(ewma(price, period), period), period)
    elif moving_average == MovingAverage.wma:
        ma3 = wma(wma(wma(price, period), period), period)

    return ma3.diff(1) / ma3.shift(1)


def trix_signal(price: Series, period: int, signal_period: int, moving_average: MovingAverage) -> Series:
    """
    트라이엄프엑스시그널

    <설명>
    트라이엄프엑스시그널(TRIX Signal)을 구하는 함수입니다.
    트라이엄프엑스시그널(TRIX Signal)은 트라이엄프엑스(TRIX)의 이동 평균 값입니다.

    <사용 방법>
    첫 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 가격을,
    두 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 기간을,
    세 번째 인자에는 트라이엄프엑스시그널(TRIX Signal)을 구하는데 사용하는 시그널 기간을,
    네 번째 인자에는 트라이엄프엑스(TRIX)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 10일간 종가의 지수 이동 평균으로 트라이엄프엑스(TRIX)를 구하고 9일간 트라이엄프엑스(TRIX)의 지수 이동 평균을 구하고자 하는 경우에는
    'trix_signal(close, 10, 9, ema)' 또는 '소나시그널(종가, 10, 9, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price: (가격데이터) 트라이엄프엑스(TRIX)를 구하는데 사용하는 가격
    :param period: (기간) 트라이엄프엑스(TRIX)를 구하는데 사용하는 기간
    :param signal_period: (시그널기간) 트라이엄프엑스시그널(TRIX Signal)을 구하는데 사용하는 시그널 기간
    :param moving_average: (이동평균종류) 트라이엄프엑스(TRIX)를 구하는데 사용하는 이동 평균 종류
    :return:
    """
    trix_val = trix(price, period, moving_average)
    if moving_average == MovingAverage.sma:
        return sma(trix_val, signal_period)
    elif moving_average == MovingAverage.ema:
        return ema(trix_val, signal_period)
    elif moving_average == MovingAverage.ewma:
        return ewma(trix_val, signal_period)
    elif moving_average == MovingAverage.wma:
        return wma(trix_val, signal_period)

def pdi(
        price_high: Series, price_low: Series, price_close: Series, period: int, moving_average: MovingAverage
) -> Series:
    """
    매수방향지표

    <설명>
    매수방향지표(PDI)를 구하는 함수입니다.
    매수방향지표(PDI)는 실질적으로 상승하는 폭의 비율을 나타냅니다.
    매수방향지표(PDI)는 0에서 1사이의 값으로 표현됩니다.

    <사용 방법>
    첫 번째 인자에는 고가를,
    두 번째 인자에는 저가를,
    세 번째 인자에는 종가를,
    네 번째 인자에는 매수방향지표(PDI)를 구하는데 사용하는 기간을,
    다섯 번째 인자에는 매수방향지표(PDI)를 구하는데 사용하는 이동 평균 종류를 적으면 됩니다.
    예를 들어, 지수 이동 평균을 이용한 14일간 매수방향지표(PDI)를 구하고자 하는 경우
    'pdi(high, low, close, 14, ema)' 또는 '매수방향지표(고가, 저가, 종가, 14, 지수이동평균)'과 같이 작성하면 됩니다.

    :param price_high: (고가) 고가
    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) 매수방향지표(PDI)를 구하는데 사용하는 기간
    :param moving_average: (이동평균종류) 매수방향지표(PDI)를 구하는데 사용하는 이동 평균 종류 ex) 단순 이동평균, 지수 이동평균, 가중 이동평균
    :return:
    """
    pdm = np.where(
        ((price_high.diff(1) > 0) & (price_high.diff(1) > price_low.shift(1) - price_low)), price_high.diff(1), 0
    )

    if moving_average == MovingAverage.sma:
        pdmn = sma(Series(pdm), period)
    elif moving_average == MovingAverage.ema:
        pdmn = ema(Series(pdm), period)
    elif moving_average == MovingAverage.ewma:
        pdmn = ewma(Series(pdm), period)
    elif moving_average == MovingAverage.wma:
        pdmn = wma(Series(pdm), period)

    tr = _tr(price_high, price_low, price_close)

    if moving_average == MovingAverage.sma:
        trn = sma(tr, period)
    elif moving_average == MovingAverage.ema:
        trn = ema(tr, period)
    elif moving_average == MovingAverage.ewma:
        trn = ewma(tr, period)
    elif moving_average == MovingAverage.wma:
        trn = wma(tr, period)

    return pdmn.divide(trn)


def RSI(price: Series, period: int):

    U = np.where(price.diff(1) > 0, price.diff(1), 0)
    D = np.where(price.diff(1) < 0, price.diff(1) * (-1), 0)

    AU = DataFrame(U).rolling(window=period, min_periods=period).mean()
    AD = DataFrame(D).rolling(window=period, min_periods=period).mean()
    RSI = AU.div(AD + AU)
    return RSI


from datetime import date, timedelta
import pandas as pd
from pandas import DataFrame
import math

from buke_pypi.indicator import sma, pivot_standard, stochastic_fast_k, stochastic_fast_d, pivot_standard_for_sellprice, ibs, RSI
from buke_pypi.function import rank, ts_min, ts_max
from buke_pypi.parameter import Market, MovingAverage
from buke_pypi.tools import get_ask_price, get_stock_exchange_tax, get_bid_price


def what_day_is_it(date):
    days = ['월요일', '화요일', '수요일', '목요일', '금요일', '토요일', '일요일']
    day = date.weekday()
    return days[day]

def str_to_date(trading_day: str) -> date:
    return date(int(trading_day[:4]), int(trading_day[5:7]), int(trading_day[8:]))


def int_to_str(x: int) -> str:

    try:
        x = int(x)
        return "%06d" % x
    except:
        return x


def read_trading_days(dir, start_date: date, end_date: date):
    df = pd.read_csv(f"{dir}/data_file/korea_trading_days.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    trading_days = df['날짜'].to_list()
    trading_days = [day for day in trading_days if start_date <= day <= end_date]
    return trading_days


def read_day_price(dir):
    df = pd.read_csv(f"{dir}/data_file/전체데이터.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    # try:
    df['종목코드'] = df['종목코드'].apply(lambda x: int_to_str(x))
    # except:
    #     pass
    return df


def read_index_day_price(dir):
    df = pd.read_csv(f"{dir}/data_file/Kosdaq_Price.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    return df


def read_index_history(dir):
    ks_history = pd.read_csv(f"{dir}/data_file/코스피_종목리스트.csv", index_col=0)
    ks_history['날짜'] = ks_history['날짜'].apply(lambda day: str_to_date(day))
    ks_history = ks_history.set_index('날짜')

    kq_history = pd.read_csv(f"{dir}/data_file/코스닥_종목리스트.csv")
    kq_history['날짜'] = kq_history['날짜'].apply(lambda day: str_to_date(day))
    kq_history = kq_history.set_index('날짜')
    return ks_history, kq_history


def add_indicator(day_price: DataFrame, index_day_price: DataFrame) -> DataFrame:

    group_list = []
    grouped = day_price.groupby('종목코드')
    for key, group in grouped:
        group = group.reset_index(drop=True)

        group['거래대금_이평선_5'] = sma(group['거래대금'], 5)

        group['pivot_기준선'] = pivot_standard(group['고가'], group['저가'], group['종가'])
        group['pivot_기준선_for_sell'] = pivot_standard_for_sellprice(group['고가'], group['저가'], group['종가'])

        group['상승비율'] = group['종가'] / group['종가'].shift(1) - 1
        group['상승비율_양수'] = group['상승비율'].apply(lambda x: max(0, x))

        group['종가_이평선_20'] = sma(group['종가'], 20)

        group['ibs'] = ibs(group['고가'], group['저가'], group['종가'])

        group['스토캐스틱_fast_10'] = stochastic_fast_k(group['고가'], group['저가'], group['종가'], 10)
        group['스토캐스틱_slow_10_3'] = stochastic_fast_d(group['고가'], group['저가'], group['종가'], 10, 3, MovingAverage.sma)

        group['변동폭_저가'] = group['종가'] / group['저가']
        group['변동폭_저가_이평선_10'] = sma(group['변동폭_저가'], 10)

        group['RSI_5'] = RSI(group['종가'], 5)
        group['RSI_5_이평선_5'] = sma(group['RSI_5'], 5)

        group['변동폭'] = group['종가'] / group['고가']
        group['변동폭_이평선_3'] = sma(group['변동폭'], 3)
        group['변동폭_이평선_10'] = sma(group['변동폭'], 10)

        group['ibs_이평선_10'] = sma(group['ibs'], 10)

        group['스토캐스틱_fast_15'] = stochastic_fast_k(group['고가'], group['저가'], group['종가'], 15)
        group['스토캐스틱_slow_15_5'] = stochastic_fast_d(group['고가'], group['저가'], group['종가'], 15, 5, MovingAverage.sma)

        group_list.append(group)

    day_price = pd.concat(group_list, axis=0)
    day_price = day_price.reset_index(drop=True)

    group_list = []
    grouped = day_price.groupby('날짜')
    for key, group in grouped:

        group = group.reset_index(drop=True)
        group['시가총액_순위'] = rank(group['시가총액'])
        group['거래대금_이평선_5_순위'] = rank(group['거래대금_이평선_5'])

        상승종목갯수 = (group['상승비율_양수'] > 0).sum()
        상승비율합 = group['상승비율_양수'].dropna().sum()
        매도가격 = 상승비율합 / 상승종목갯수

        group['매도가격'] = 매도가격

        group_list.append(group)

    day_price = pd.concat(group_list, axis=0)
    day_price = day_price.reset_index(drop=True)

    day_price_dict = {}
    grouped = day_price.groupby('종목코드')
    for key, group in grouped:

        increase_condition1 = group['종가'] / group['종가_이평선_20'] >  1.05
        increase_condition2 = group['변동폭_저가'] / group['변동폭_저가_이평선_10'] <  0.99
        increase_condition3 = group['RSI_5'] / group['RSI_5_이평선_5'] <  0.97
        increase_condition4 = group['변동폭_이평선_3'] / group['변동폭_이평선_10'] < 1.0
        increase_condition5 = group['ibs_이평선_10'] >  0.4
        increase_condition6 = group['스토캐스틱_fast_10'] / group['스토캐스틱_slow_10_3'] <  0.9
        increase_condition7 = group['ibs_이평선_10'] - group['ibs'] > 0.2
        increase_condition8 = group['변동폭_저가'] <  1.01

        increase_condition = increase_condition1 & increase_condition2 & increase_condition3 & increase_condition4 & \
                             increase_condition5 & increase_condition6 & increase_condition7 & increase_condition8

        liquidity_condition1 = group['시가총액_순위'] > 0.5
        liquidity_condition2 = group['거래대금_이평선_5_순위'] > 0.5

        liquidity_condition = liquidity_condition1 & liquidity_condition2

        # Result
        group['#final_result'] = increase_condition & liquidity_condition

        group['#priority'] = -1 * (group['스토캐스틱_fast_15'] / group['스토캐스틱_slow_15_5'])
        group = group.set_index('날짜')
        day_price_dict[key] = group

    # Market Timing
    index_day_price['#market_timing'] = True

    index_day_price = index_day_price.set_index('날짜')

    return day_price_dict, index_day_price


def back_testing(dir, start_date, end_date):
    # 0. 변수
    max_basket_size = 10
    liquidation_holding_days = 1
    close_holding_days = 2
    cut_period = 0
    seed_money = 100000000
    fee = 0.00015  # when using kiwoom
    basket = {}  # key:ticker, value:buy_price, stock_num, holding_days
    pnl = {'date': [start_date], 'balance': [seed_money]}
    p_len = []

    # 1. 거래일 가져오기
    trading_days_list = read_trading_days(dir, start_date, end_date)

    # 2. 일봉 가져오기
    day_price = read_day_price(dir)

    index_day_price = read_index_day_price(dir)

    # 3. 지수 종목 구성 내역 가져오기
    kospi_history, kosdaq_history = read_index_history(dir)

    # 4. 매수 조건 및 우선순위 생성
    day_price, index_day_price = add_indicator(day_price, index_day_price)
    print(f"Finished data preprocessing...")
    # print(day_price[['#final_result','#priority','market_timing']])
    # print(day_price.tail())

    plus_df = {'name': [], 'ticker': [], 'buy_date': [], 'sell_date': [], 'buy_price': [], 'sell_price': [],
               'profit': []}
    minus_df = {'name': [], 'ticker': [], 'buy_date': [], 'sell_date': [], 'buy_price': [], 'sell_price': [],
                'profit': []}

    # 5. 백테스트
    for i, day in enumerate(trading_days_list):

        if i > 1:
            # 당일 코스피, 코스닥 구성 종목 내역 가져오기
            kospi_ticker_list = kospi_history['종목리스트'].loc[day].split(',')
            kosdaq_ticker_list = kosdaq_history['종목리스트'].loc[day].split(',')

            # Market Timing
            buy_check = 1
            try:
                yesterday_market_time = bool(index_day_price['#market_timing'].loc[trading_days_list[i - 1]])
            except:
                buy_check = 0

            # 매수
            if len(basket) < max_basket_size and buy_check == 1:
                # buy at open price
                pre_buy_list = []
                for ticker in kospi_ticker_list + kosdaq_ticker_list:

                    if ticker in basket:
                        continue
                    try:
                        yesterday_buy_condition = bool(
                            day_price[ticker]['#final_result'].loc[trading_days_list[i - 1]])

                        tomorrow_vol = int(day_price[ticker]['거래량'].loc[day])
                        vol_condition = tomorrow_vol > 0

                        if yesterday_market_time and yesterday_buy_condition :
                            priority_value = float(day_price[ticker]['#priority'].loc[trading_days_list[i - 1]])
                            ticker_data = (priority_value, ticker)
                            pre_buy_list.append(ticker_data)
                    except:
                        pass

                # 여기도 수정을 한번 해봄
                if len(pre_buy_list) > 0:

                    p_len.append(len(pre_buy_list))
                    pre_buy_list.sort(reverse=True)
                    print(f"{day}, {pre_buy_list}")

                    for priority_value, ticker in pre_buy_list[:max_basket_size - len(basket)]:

                        try:
                            yesterday_low = int(day_price[ticker]['저가'].loc[trading_days_list[i - 1]])
                            yesterday_close = int(day_price[ticker]['종가'].loc[trading_days_list[i - 1]])
                            yesterday_mid = (yesterday_low + yesterday_close) / 2


                            if ticker in kospi_ticker_list:
                                today_target_buy_price = get_ask_price(yesterday_mid * 0.98, Market.kospi, day)
                            else:
                                today_target_buy_price = get_ask_price(yesterday_mid * 0.98, Market.kosdaq, day)

                            today_open = int(day_price[ticker]['시가'].loc[day])
                            today_low = int(day_price[ticker]['저가'].loc[day])
                            today_name = day_price[ticker]['종목명'].loc[day]

                            real_buy_price = 0.

                            if today_open <= today_target_buy_price:
                                buy_price = math.ceil(today_open * (1 + fee))
                                real_buy_price = today_open

                            elif today_low < today_target_buy_price:
                                buy_price = math.ceil(today_target_buy_price * (1 + fee))
                                real_buy_price = today_target_buy_price
                            else:
                                continue

                            if ticker in kospi_ticker_list:
                                sell_price = get_ask_price(buy_price * 1.05, Market.kospi, day)

                            else:
                                sell_price = get_ask_price(buy_price * 1.05, Market.kosdaq, day)


                            stock_num = (pnl['balance'][-1] // max_basket_size) // buy_price
                            stock_info = {'stock_num': stock_num, 'buy_price': buy_price, 'holding_days': 0,
                                          'name': today_name, 'buy_date': day, 'real_buy_price': real_buy_price, 'sell_price': sell_price}

                            basket[ticker] = stock_info
                            print(f"<Buy> date: {day}, ticker: {ticker}, buy_price: {buy_price}, ")
                        except:
                            print(f"pnl:{pnl['balance'][-1]} max_basket_size:{max_basket_size} buy_price:{buy_price}")


            # 매도
            today_profit = 0
            if len(basket) > 0:
                tax = get_stock_exchange_tax(day)

                sell_list = []
                for ticker in basket:

                    holding_condition1 = basket[ticker]['holding_days'] >= liquidation_holding_days

                    if holding_condition1:

                        try:

                            today_open = int(day_price[ticker]['시가'].loc[day])
                            today_high = int(day_price[ticker]['고가'].loc[day])

                            tomorrow_pivot_standard_price = float(day_price[ticker]['pivot_기준선_for_sell'].loc[day])
                            tomorrow_sell_price = float(day_price[ticker]['매도가격'].loc[day])

                            buy_date = basket[ticker]['buy_date']
                            sell_date = day
                            name = basket[ticker]['name']

                            target_sell_price = float(basket[ticker]['sell_price'])

                            if ticker in kospi_ticker_list:
                                tomorrow_target_sell_price = get_bid_price(tomorrow_pivot_standard_price * (1 + tomorrow_sell_price/3),
                                                                  Market.kospi, day)
                            else:
                                tomorrow_target_sell_price = get_bid_price(tomorrow_pivot_standard_price * (1 + tomorrow_sell_price/3),
                                                                  Market.kosdaq, day)

                            basket[ticker]['sell_price'] = tomorrow_target_sell_price


                            if today_open >= target_sell_price:
                                sell_price = today_open
                            elif today_high > target_sell_price:
                                sell_price = target_sell_price
                            else:
                                basket[ticker]['holding_days'] += 1

                                # print(f"<UnSell> date: {day}, name: {basket[ticker]['name']}({ticker}), tomorrow_sell_price: {tomorrow_target_sell_price}")
                                print(f"<UnSell> date: {day}, name: {basket[ticker]['name']}({ticker})")
                                continue

                            profit = basket[ticker]['stock_num'] * (
                                    math.floor(sell_price * (1 - (tax + fee))) - float(basket[ticker]['buy_price']))
                            today_profit += profit

                            if profit > 0:
                                plus_df['buy_date'].append(buy_date)
                                plus_df['sell_date'].append(sell_date)
                                plus_df['ticker'].append(ticker)
                                plus_df['profit'].append(profit)
                                plus_df['name'].append(name)
                                plus_df['buy_price'].append(basket[ticker]['buy_price'])
                                plus_df['sell_price'].append(sell_price)

                            else:
                                minus_df['buy_date'].append(buy_date)
                                minus_df['sell_date'].append(sell_date)
                                minus_df['ticker'].append(ticker)
                                minus_df['profit'].append(profit)
                                minus_df['name'].append(name)
                                minus_df['buy_price'].append(basket[ticker]['buy_price'])
                                minus_df['sell_price'].append(sell_price)

                            avg = sell_price / basket[ticker]['buy_price']
                            avg = round((1 - avg) * 100, 2)
                            print(f"<Sell> date: {day}, name: {basket[ticker]['name']}({ticker}), sell_price: {sell_price}, profit: {avg}%")
                            sell_list.append(ticker)

                        except:
                            print(f"<Except> date: {day} ")
                            basket[ticker]['holding_days'] += 1
                            continue
                    else:
                        basket[ticker]['holding_days'] += 1

                for ticker in sell_list:
                    basket.pop(ticker)

            # Balance Update
            pnl['date'].append(day)
            pnl['balance'].append(pnl['balance'][-1] + today_profit)

    # print(sum(p_len) / len(p_len))
    return basket

def Buke_ST004_sell_list(dir, today):

    path = f"{dir}/trade_list/{today}"

    end_date = today  # today는 토요일 or 일요일
    start_date = end_date - timedelta(days=30)

    print(start_date, end_date)
    trading_days_list = read_trading_days(dir, start_date, end_date)

    real_start_date = trading_days_list[0]
    real_end_date = trading_days_list[-1]
    print("start_date : ", what_day_is_it(real_start_date))
    print("end_date   : ", what_day_is_it(real_end_date))
    print()

    basket = back_testing(dir, real_start_date, real_end_date)

    file_name = "sell_4"
    sell_basket = {'종목코드': [], '종목명': [], '매도가격': []}

    for ticker in basket:
        sell_price = basket[ticker]['sell_price']
        name = basket[ticker]['name']

        sell_basket['종목코드'].append(ticker)
        sell_basket['종목명'].append(name)
        sell_basket['매도가격'].append(sell_price)

    sell_basket = pd.DataFrame(sell_basket)
    sell_basket.to_csv(f"{path}/{file_name}_{str(today)[-5:]}.txt")


from pandas import Series, DataFrame, merge
import numpy as np
# from sklearn.linear_model import LinearRegression

from buke_pypi.parameter import KindOfAverage

#
# def _linear_regression(x, y):
#     linear = LinearRegression()
#     model = linear.fit(x, y)
#     coef = model.coef_
#     return coef


def s_max(*args) -> Series:
    """
    최댓값

    <설명>
    입력한 인자들 가운데 가장 큰 값을 반환하는 함수입니다.

    <사용 방법>
    비교할 값들을 인자로 적으면 됩니다.
    예를 들어, 시가와 종가 중 큰 값만을 사용하고자 하는 경우에는
    's_max(open, close)' 또는 '최댓값(시가, 종가)'와 같이 작성하면 됩니다.

    :param args:
    :return:
    """
    series_list = []
    for obj in args:
        series_list.append(obj)

    merged_data = DataFrame(series_list)
    return merged_data.max(axis=0)


def s_min(*args) -> Series:
    """
    최솟값

    <설명>
    입력한 인자들 가운데 가장 작은 값을 반환하는 함수입니다.

    <사용 방법>
    비교할 값들을 인자로 적으면 됩니다.
    예를 들어, 시가와 종가 중 작은 값만을 사용하고자 하는 경우에는
    's_min(open, close)' 또는 '최솟값(시가, 종가)'와 같이 작성하면 됩니다.

    :param args:
    :return:
    """
    series_list = []
    for obj in args:
        series_list.append(obj)

    merged_data = DataFrame(series_list)
    return merged_data.min(axis=0)


def rank(value: Series) -> Series:
    """
    순위

    <설명>
    당일 코스피, 코스닥 구성 종목에서 해당 종목의 순위를 반환하는 함수입니다.
    0과 1사이의 값을 가지며, 1에 가까울수록 순위가 높다는 의미입니다.

    <사용 방법>
    첫 번째 인자에 순위를 구하고자 하는 값을 적으면 됩니다.
    예를 들어, 20일 평균 거래대금의 순위를 구하고자 하는 경우에는
    'rank(sma(tr_val, 20))' 또는 '순위(단순이동평균(거래대금, 20))'과 같이 작성하면 됩니다.

    :param value: (데이터) 순위를 구하고자 하는 값
    :return:
    """
    return value.rank(pct=True)


def ts_sum(value: Series, period: int) -> Series:
    """
    기간합계

    <설명>
    기간합계(ts_sum)는 특정 기간(period)동안 값(value)들의 합을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간합계(ts_sum)를 구하고자 하는 값을,
    두 번째 인자는 기간합계(ts_sum)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 5일간 거래량의 합을 구하고자 하는 경우에는
    'ts_sum(volume, 5)' 또는 '기간합계(거래량, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간합계(ts_sum)를 구하고자 하는 값
    :param period: (기간) 기간합계(ts_sum)를 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).sum()


def ts_max(value: Series, period: int) -> Series:
    """
    기간최댓값

    <설명>
    기간최댓값(ts_max)은 특정 기간(period)동안 값(value)들 중 최댓값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간최댓값(ts_max)을 구하고자 하는 값을,
    두 번째 인자는 기간최댓값(ts_max)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일간 종가 중 최댓값을 구하고자 하는 경우에는
    'ts_max(close, 5)' 또는 '기간최댓값(종가, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간최댓값(ts_max)을 구하고자 하는 값
    :param period: (기간) 기간최댓값(ts_max)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).max()


def ts_min(value: Series, period: int) -> Series:
    """
    기간최솟값

    <설명>
    기간최솟값(ts_min)은 특정 기간(period)동안 값(value)들 중 최솟값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간최솟값(ts_min)을 구하고자 하는 값을,
    두 번째 인자는 기간최솟값(ts_min)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일간 종가 중 최솟값을 구하고자 하는 경우에는
    'ts_min(close, 5)' 또는 '기간최솟값(종가, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간최솟값(ts_min)을 구하고자 하는 값
    :param period: (기간) 기간최솟값(ts_min)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).min()


def ts_median(value: Series, period: int) -> Series:
    """
    기간중앙값

    <설명>
    기간중앙값(ts_median)은 특정 기간(period)동안 값(value)들 중 중앙값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간중앙값(ts_median)을 구하고자 하는 값을,
    두 번째 인자는 기간중앙값(ts_median)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일간 종가 중에서 중앙값을 구하고자 하는 경우에는
    'ts_median(close, 5)' 또는 '기간중앙값(종가, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간중앙값(ts_median)을 구하고자 하는 값
    :param period: (기간) 기간중앙값(ts_median)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).median()


def ts_argmax(value: Series, period: int) -> Series:
    """
    기간최대일

    <설명>
    기간최대일(ts_argmax)은 특정 기간(period)동안 값(value)들 중 최댓값이 되는 날을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간최대일(ts_argmax)을 구하고자 하는 값을,
    두 번째 인자는 기간최대일(ts_argmax)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일간 종가의 기간최대일(ts_argmax)을 구하고자 하는 경우에는
    'ts_argmax(close, 5)' 또는 '기간최대일(종가, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간최대일(ts_argmax)을 구하고자 하는 값
    :param period: (기간) 기간최대일(ts_argmax)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).apply(lambda x: Series(x).argmax())


def ts_argmin(value: Series, period: int) -> Series:
    """
    기간최소일

    <설명>
    기간최소일(ts_argmin)은 특정 기간(period)동안 값(value)들 중 최솟값이 되는 날을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간최소일(ts_argmin)을 구하고자 하는 값을,
    두 번째 인자는 기간최소일(ts_argmin)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일간 종가의 기간최소일(ts_argmin)을 구하고자 하는 경우에는
    'ts_argmin(close, 5)' 또는 '기간최소일(종가, 5)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간최소일(ts_argmin)을 구하고자 하는 값
    :param period: (기간) 기간최소일(ts_argmin)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).apply(lambda x: Series(x).argmin())


def ts_quantile(value: Series, period: int, percentile: float) -> Series:
    """
    기간백분위수

    <설명>
    기간백분위수(ts_quantile)는 특정 기간(period)동안의 값(value)들 중 특정 백분위(percentile)에 해당하는 값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간백분위수(ts_quantile)을 구하고자 하는 값을,
    두 번째 인자는 기간백분위수(ts_quantile)을 구하는데 사용하고자 하는 기간,
    세 번째 인자는 기간백분위수(ts_quantile)을 구하는데 사용하는 백분위를 적으면 됩니다.
    예를 들어, 최근 20일간 종가 중에서 75%에 해당하는 값을 구하고자 하는 경우에는
    'ts_quantile(close, 20, 0.75)' 또는 '기간백분위수(종가, 20, 0.75)'와 같이 작성하면 됩니다.

    :param value: (데이터) 기간백분위수(ts_quantile)를 구하고자 하는 값
    :param period: (기간) 기간백분위수(ts_quantile)를 구하는데 사용하고자 하는 기간
    :param percentile: (백분위값) 기간백분위수(ts_quantile)를 구하는데 사용하는 백분위 값
    :return:
    """
    return value.rolling(window=period, min_periods=period).quantile(quantile=percentile)


def ts_stddev(value: Series, period: int) -> Series:
    """
    기간표준편차

    <설명>
    기간표준편차(ts_stddev)는 특정 기간(period)동안 값(value)들의 표준편차를 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간표준편차(ts_stddev)를 구하고자 하는 값을,
    두 번째 인자는 기간표준편차(ts_stddev)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 20일간 종가의 표준편차를 구하고자 하는 경우에는
    'ts_stddev(close, 20)' 또는 '기간표준편차(종가, 20)'과 같이 작성하면 됩니다.

    :param value: (데이터) 기간표준편차(ts_stddev)를 구하고자 하는 값
    :param period: (기간) 기간표준편차(ts_stddev)를 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period).std(ddof=1)


def pct_change(value: Series, period: int) -> Series:
    """
    변화율

    <설명>
    변화율(pct_change)은 특정 기간(period)동안 값(value)의 변화율을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 변화율(pct_change)을 구하고자 하는 값을,
    두 번째 인자는 변화율(pct_change)을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 20일간 종가의 변화율 즉 20일 모멘텀을 구하고자 하는 경우에는
    'pct_change(close, 20)' 또는 '변화율(종가, 20)'과 같이 작성하면 됩니다.

    :param value: (데이터) 변화율(pct_change)을 구하고자 하는 값
    :param period: (기간) 변화율(pct_change)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.pct_change(periods=period)
    # 정확한 시간 측정을 통해 빠른 연산을 선택해야
    # return value / value.shift(period) - 1


def sigmoid(value: Series) -> Series:
    """
    시그모이드

    <설명>
    값들을 정규화하는 함수 중 하나입니다.
    시그모이드(sigmoid)를 사용하면 값들을 0과 1사이로 정규화시킬 수 있습니다.

    <사용 방법>
    첫 번째 인자에 정규화시키고자하는 값을 적으면 됩니다.
    예를 들어, 5일 종가 평균을 시그모이드 함수를 이용해 0과 1사이의 값으로 만들고자하는 경우에는
    'sigmoid(sma(close, 5))' 또는 '시그모이드(단순이동평균(종가, 5))'와 같이 작성하면 됩니다.

    :param value: (데이터) 정규화시키고자하는 값
    :return:
    """
    return 1 / (1 + np.exp(-value))


def sine(value: Series) -> Series:
    """
    사인

    <설명>
    입력된 값을 radian 단위로 취급하여 사인 값을 구하는 함수입니다.
    사인(sine)을 사용하면 값들을 -1과 1사이로 정규화시킬 수 있습니다.

    <사용 방법>
    첫 번째 인자에 사인 값으로 반환할 대상을 적으면 됩니다.
    예를 들어, 20일 종가의 변화율을 사인 값으로 변환하고자 하는 경우에는
    'sine(pct_change(close, 20))' 또는 '사인(변화율(종가, 20))'과 같이 작성하면 됩니다.

    :param value: (데이터) 사인 값으로 변환하고자 하는 값
    :return:
    """
    return np.sin(value)


def cosine(value: Series) -> Series:
    """
    코사인

    <설명>
    입력된 값을 radian 단위로 취급하여 코사인 값을 구하는 함수입니다.
    코사인(cosine)을 사용하면 값들을 -1과 1사이로 정규화시킬 수 있습니다.

    <사용 방법>
    첫 번째 인자에 코사인 값으로 반환할 대상을 적으면 됩니다.
    예를 들어, 20일 종가의 변화율을 코사인 값으로 변환하고자 하는 경우에는
    'cosine(pct_change(close, 20))' 또는 '코사인(변화율(종가, 20))'과 같이 작성하면 됩니다.

    :param value: (데이터) 코사인 값으로 변환하고자 하는 값
    :return:
    """
    return np.cos(value)


def tangent(value: Series) -> Series:
    """
    탄젠트

    <설명>
    입력된 값을 radian 단위로 취급하여 탄젠트 값을 구하는 함수입니다.
    탄젠트(tangent)을 사용하면 값들을 -1과 1사이로 정규화시킬 수 있습니다.

    <사용 방법>
    첫 번째 인자에 탄젠트 값으로 반환할 대상을 적으면 됩니다.
    예를 들어, 20일 종가의 변화율을 탄젠트 값으로 변환하고자 하는 경우에는
    'tangent(pct_change(close, 20))' 또는 '탄젠트(변화율(종가, 20))'과 같이 작성하면 됩니다.

    :param value: (데이터) 탄젠트 값으로 변환하고자 하는 값
    :return:
    """
    return np.tan(value)


def tanh(value: Series) -> Series:
    """
    하이퍼블릭탄젠트

    <설명>
    값들을 정규화하는 함수 중 하나입니다.
    하이퍼블릭탄젠트(tanh)를 사용하면 값들을 0과 1사이로 정규화시킬 수 있습니다.

    <사용 방법>
    첫 번째 인자에 정규화시키고자하는 값을 적으면 됩니다.
    예를 들어, 5일 종가 평균을 시그모이드 함수를 이용해 0과 1사이의 값으로 만들고자하는 경우에는
    'sigmoid(sma(close, 5))' 또는 '시그모이드(단순이동평균(종가, 5))'와 같이 작성하면 됩니다.

    :param value: (데이터) 정규화시키고자하는 값
    :return:
    """
    return np.tanh(value)


def ts_rank(value: Series, period: int) -> Series:
    """
    기간순위

    <설명>
    한 종목의 당일 값의 순위를 특정 기간동안의 값들 중에서 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간순위(ts_rank)를 구하고자 하는 값을,
    두 번째 인자는 기간순위(ts_rank)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 60일간 종가의 기간순위를 구하고자 하는 경우에는
    'ts_rank(close, 60)' 또는 '기간순위(종가, 60)'과 같이 작성하면 됩니다.

    :param value: (데이터) 기간순위(ts_rank)를 구하고자 하는 값
    :param period: (기간) 기간순위(ts_rank)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).apply(
        lambda x: Series(x).rank(pct=True, method=KindOfAverage.average.name).iloc[-1]
    )


def ts_zscore(value: Series, period: int) -> Series:
    """
    기간표준점수

    <설명>
    한 종목의 당일 표준점수를 특정 기간동안의 값들을 이용해서 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간표준점수(ts_zscore)를 구하고자 하는 값을,
    두 번째 인자는 기간표준점수(ts_zscore)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 60일간 종가의 기간표준점수(ts_zscore)를 구하고자 하는 경우에는
    'ts_zscore(close, 60)' 또는 '기간표준점수(종가, 60)'과 같이 작성하면 됩니다.

    :param value: (데이터) 기간표준점수(ts_zscore)를 구하고자 하는 값
    :param period: (기간) 기간표준점수(ts_zscore)을 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).apply(
        lambda x: (Series(x).iloc[-1] - Series(x).mean()) / Series(x).std(ddof=1)
    )


def ts_skew(value: Series, period: int) -> Series:
    """
    기간왜도

    <설명>
    기간왜도(ts_skew)는 특정 기간(period)동안 값(value)들의 왜도(skewness) 값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간왜도(ts_skew)를 구하고자 하는 값을,
    두 번째 인자는 기간왜도(ts_skew)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 20일간 종가의 왜도 값을 구하고자 하는 경우에는
    'ts_skew(close, 20)' 또는 '기간왜도(종가, 20)'과 같이 작성하면 됩니다.

    :param value: (데이터) 기간왜도(ts_skew)를 구하고자 하는 값
    :param period: (기간) 기간왜도(ts_skew)를 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).skew()


def ts_kurt(value: Series, period: int) -> Series:
    """
    기간첨도

    <설명>
    기간첨도(ts_kurt)는 특정 기간(period)동안 값(value)들의 첨도(kurtosis) 값을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 기간첨도(ts_kurt)를 구하고자 하는 값을,
    두 번째 인자는 기간첨도(ts_kurt)를 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 20일간 종가의 왜도 값을 구하고자 하는 경우에는
    'ts_kurt(close, 20)' 또는 '기간첨도(종가, 20)'과 같이 작성하면 됩니다.

    :param value: (데이터) 기간첨도(ts_kurt)를 구하고자 하는 값
    :param period: (기간) 기간첨도(ts_kurt)를 구하는데 사용하고자 하는 기간
    :return:
    """
    return value.rolling(window=period, min_periods=period).kurt()


def increase_from_lowest_price(price_low: Series, price_close: Series, period: int) -> Series:
    """
    최저가대비상승

    <설명>
    특정 기간(period)동안의 최저가 대비 당일 종가의 상승폭을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 저가를,
    두 번째 인자는 종가를,
    세 번째 인자는 최저가대비상승을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일 동안의 최저가 대비 상승을 구하고자 하는 경우에는
    'increase_from_lowest_price(low, close, 5) 또는 '최저가대비상승(저가, 종가, 5)'와 같이 작성하면 됩니다.

    :param price_low: (저가) 저가
    :param price_close: (종가) 종가
    :param period: (기간) 최저가대비상승(increase_from_lowest_price)을 구하는데 사용하고자 하는 기간
    :return:
    """
    lowest_price = ts_min(price_low, period)
    return price_close / lowest_price - 1


def decrease_from_highest_price(price_high: Series, price_close: Series, period: int) -> Series:
    """
    최고가대비하락

    <설명>
    특정 기간(period)동안의 최고가 대비 당일 종가의 하락폭을 구하는 함수입니다.

    <사용 방법>
    첫 번째 인자는 고가를,
    두 번째 인자는 종가를,
    세 번째 인자는 최고가대비하락을 구하는데 사용하고자 하는 기간을 적으면 됩니다.
    예를 들어, 최근 5일 동안의 최고가 대비 하락을 구하고자 하는 경우에는
    'decrease_from_highest_price(high, close, 5) 또는 '최고가대비하락(고가, 종가, 5)'와 같이 작성하면 됩니다.

    :param price_high: (저가) 고가
    :param price_close: (종가) 종가
    :param period: (기간) 최고가대비하락(decrease_from_highest_price)을 구하는데 사용하고자 하는 기간
    :return:
    """
    highest_price = ts_max(price_high, period)
    return price_close / highest_price - 1


def ts_regression(value: Series, period: int) -> Series:
    return value.rolling(window=period, min_periods=period).apply(
        lambda x: _linear_regression(
            [[x_val] for x_val in np.arange(1, period + 1)],
            [[(y_val - Series(x).values[0]) / Series(x).values[0] * 100] for y_val in Series(x).values],
        )
    )


def get_ts_correlation(lhs_value: DataFrame, rhs_value: DataFrame, window_size: int):
    merged = merge(left=lhs_value, right=rhs_value, how="outer", on="date", sort=True, suffixes=["-lhs", "-rhs"])
    total_value = DataFrame(merged["date"])

    lhs_roll = merged["close-lhs"].rolling(window_size)
    rhs_roll = merged["close-rhs"].rolling(window_size)
    rhs_iterator = iter(rhs_roll)
    correlation_list = list()
    for lhs in lhs_roll:
        rhs = next(rhs_iterator)
        correlation_list.append(lhs.corr(rhs))

    total_value["correlation"] = correlation_list

    lhs = merge(lhs_value, total_value, on="date")
    rhs = merge(rhs_value, total_value, on="date")
    return lhs, rhs



from datetime import date
from math import floor

from buke_pypi.parameter import Market


def year_to_date(year: int):
    return date(year=year, month=1, day=1)


def get_hoga(price: int or float, market: Market, day: date) -> int:
    """
    주가에 따른 1호가 가격을 구하는 함수
    :param price: 주가
    :param market: kospi / kosdaq
    :return: 1호가 가격
    """
    if day < date(2023,1,25):

        if market == Market.kospi:
            if price >= 500000:
                hoga = 1000
            elif price >= 100000:
                hoga = 500
            elif price >= 50000:
                hoga = 100
            elif price >= 10000:
                hoga = 50
            elif price >= 5000:
                hoga = 10
            elif price >= 1000:
                hoga = 5
            else:
                hoga = 1

        elif market == Market.kosdaq:
            if price >= 50000:
                hoga = 100
            elif price >= 10000:
                hoga = 50
            elif price >= 5000:
                hoga = 10
            elif price >= 1000:
                hoga = 5
            else:
                hoga = 1

        elif market == Market.etf:
            hoga = 5
        else:
            raise ValueError("market should be kospi or kosdaq")

    else :

        if market == Market.kospi or market == Market.kosdaq:
            if price >= 500000:
                hoga = 1000
            elif price >= 200000:
                hoga = 500
            elif price >= 50000:
                hoga = 100
            elif price >= 20000:
                hoga = 50
            elif price >= 5000:
                hoga = 10
            elif price >= 2000:
                hoga = 5
            else:
                hoga = 1

        elif market == Market.etf:
            hoga = 5

        else:
            raise ValueError("market should be kospi or kosdaq")

    return hoga


def get_bid_price(price: int or float, market: Market, day: date) -> int:
    """
    살 가격을 구하는 함수
    :param price: 현재가
    :param market: 시장
    :return: 매수 가격
    """
    hoga = get_hoga(floor(price), market, day)
    return int(price // hoga) * hoga + hoga


def get_ask_price(price: int or float, market: Market, day: date) -> int:
    """
    팔 가격을 구하는 함수
    :param price: 현재가
    :param market: 시장
    :return: 매도 가격
    """
    hoga = get_hoga(floor(price), market, day)
    if price % hoga == 0:
        return int(price // hoga) * hoga - hoga
    else:
        return int(price // hoga) * hoga


def get_limit_percent(today: date) -> float:
    if today < date(2015, 6, 15):
        return 0.15
    else:
        return 0.3


def get_upper_limit_price(yesterday_close: int, today: date, market: Market) -> int:
    """
    당일 상한가를 반환하는 함수
    :param yesterday_close: 전일 종가
    :param today: 당일 날짜
    :param market: kospi 또는 kosdaq
    :return: 당일 상한가
    """
    limit_percent = get_limit_percent(today)
    predict_upper_limit_price = yesterday_close * (1 + limit_percent)
    hoga = get_hoga(floor(predict_upper_limit_price), market, today)
    real_upper_limit_price = int(predict_upper_limit_price // hoga) * hoga

    return real_upper_limit_price


def get_lower_limit_price(yesterday_close: int, today: date, market: Market) -> int:
    """
    당일 하한가를 반환하는 함수
    :param yesterday_close: 전일 종가
    :param today: 당일 날짜
    :param market: kospi 또는 kosdaq
    :return: 당일 하한가
    """
    limit_percent = get_limit_percent(today)
    predict_lower_limit_price = yesterday_close * (1 - limit_percent)
    hoga = get_hoga(floor(predict_lower_limit_price), market, today)
    real_lower_limit_price = int(predict_lower_limit_price // hoga) * hoga + hoga

    return real_lower_limit_price


def get_stock_exchange_tax(day: date):
    """
    당일의 증권거래세를 반환하는 함수
    :param day: 당일 날짜
    :return: 당일의 증권거래세
    """
    if day < date(2019, 6, 3):
        stock_exchange_tax = 0.003
    elif day < date(2021, 1, 1):
        stock_exchange_tax = 0.0025
    elif day < date(2023, 1, 1):
        stock_exchange_tax = 0.0023
    elif day < date(2024, 1, 1):
        stock_exchange_tax = 0.002
    elif day < date(2025, 1, 1):
        stock_exchange_tax = 0.0018
    else:
        stock_exchange_tax = 0.0015
    return stock_exchange_tax



from datetime import date
import pandas as pd
from pandas import DataFrame

from buke_pypi.indicator import sma, ibs, pivot_standard, stochastic_fast_d, stochastic_fast_k, 순거래대금
from buke_pypi.function import rank
from buke_pypi.parameter import Market, MovingAverage
from buke_pypi.tools import get_ask_price


def str_to_date(trading_day: str) -> date:
    return date(int(trading_day[:4]), int(trading_day[5:7]), int(trading_day[8:]))


def int_to_str(x: int) -> str:
    return "%06d" % x


def read_trading_days(dir, end_date: date):

    df = pd.read_csv(f"{dir}/data_file/korea_trading_days.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    trading_days = df['날짜'].to_list()
    trading_days = [day for day in trading_days if day <= end_date]

    return trading_days


def read_day_price(dir, end_date: date):

    df = pd.read_csv(f"{dir}/data_file/수급데이터_2022_to_{end_date.year}.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    # df['종목코드'] = df['종목코드'].apply(lambda x: int_to_str(x))
    return df


def read_index_day_price(dir):

    df = pd.read_csv(f"{dir}/data_file/MSCI_한국_ETF.csv", index_col=0)
    df['날짜'] = df['날짜'].apply(lambda day: str_to_date(day))
    return df


def read_index_history(dir, end_date: date):

    ks_history = pd.read_csv(f"{dir}/data_file/코스피_종목리스트_2022_to_{end_date.year}.csv", index_col=0)
    ks_history['날짜'] = ks_history['날짜'].apply(lambda day: str_to_date(day))
    ks_history = ks_history.set_index('날짜')

    kq_history = pd.read_csv(f"{dir}/data_file/코스닥_종목리스트_2022_to_{end_date.year}.csv")
    kq_history['날짜'] = kq_history['날짜'].apply(lambda day: str_to_date(day))
    kq_history = kq_history.set_index('날짜')
    return ks_history, kq_history


def add_indicator(day_price: DataFrame, index_day_price: DataFrame) -> DataFrame:

    group_list = []
    grouped = day_price.groupby('종목코드')
    for key, group in grouped:
        group = group.reset_index(drop=True)
        group['거래대금_이평선_5'] = sma(group['거래대금'], 5)
        group['pivot_기준선'] = pivot_standard(group['고가'], group['저가'], group['종가'])

        group['순거래대금'] = 순거래대금(group['종가'], group['거래량'], 1)
        group['순거래대금_이평선_5'] = sma(group['순거래대금'], 5)

        group['종가_이평선_3'] = sma(group['종가'], 3)
        group['종가_이평선_5'] = sma(group['종가'], 5)
        group['종가_이평선_20'] = sma(group['종가'], 20)
        group['거래대금_이평선_10'] = sma(group['거래대금'], 10)

        group['스토캐스틱_fast'] = stochastic_fast_k(group['고가'], group['저가'], group['종가'], 20)
        group['스토캐스틱_slow'] = stochastic_fast_d(group['고가'], group['저가'], group['종가'], 20,
                                                     5, MovingAverage.sma)
        group['ibs'] = ibs(group['고가'], group['저가'], group['종가'])

        group_list.append(group)
    day_price = pd.concat(group_list, axis=0)
    day_price = day_price.reset_index(drop=True)

    group_list = []
    grouped = day_price.groupby('날짜')
    for key, group in grouped:
        group = group.reset_index(drop=True)
        group['거래대금_이평선_5_순위'] = rank(group['거래대금_이평선_5'])
        group['순거래대금_이평선_5_순위'] = rank(group['순거래대금_이평선_5'])
        group['시가총액_순위'] = rank(group['시가총액'])

        group_list.append(group)
    day_price = pd.concat(group_list, axis=0)
    day_price = day_price.reset_index(drop=True)

    day_price_dict = {}
    grouped = day_price.groupby('종목코드')
    for key, group in grouped:

        increase_condition1 = group['종가_이평선_5'] > group['종가_이평선_5'].shift(5)
        increase_condition2 = group['거래대금_이평선_10'] > group['거래대금_이평선_10'].shift(10)
        increase_condition3 = group['종가'] / group['종가_이평선_3'] < 1
        increase_condition4 = group['종가'] / group['종가_이평선_20'] > 1.05
        increase_condition5 = group['ibs'] < 0.7

        increase_condition = increase_condition1 & increase_condition2 & increase_condition3 & increase_condition4 & \
                             increase_condition5

        liquidity_condition1 = group['거래대금_이평선_5_순위'] > 0.5
        liquidity_condition2 = group['시가총액_순위'] > 0.3
        liquidity_condition3 = group['순거래대금_이평선_5_순위'] > 0.5

        liquidity_condition = liquidity_condition1 & liquidity_condition2 & liquidity_condition3

        # Result
        group['#final_result'] = increase_condition & liquidity_condition

        group['#priority'] = group['스토캐스틱_slow'] - group['스토캐스틱_fast']

        group = group.set_index('날짜')
        day_price_dict[key] = group

    # Market Timing
    index_day_price['이평선_3'] = sma(index_day_price['종가'], 3)
    index_day_price['이평선_5'] = sma(index_day_price['종가'], 5)
    index_day_price['이평선_10'] = sma(index_day_price['종가'], 10)
    # index_day_price['pct_change_21'] = pct_change(index_day_price['close'], 21)
    index_day_price['#market_timing'] = (index_day_price['종가'] > index_day_price['이평선_3']) | \
                                        (index_day_price['종가'] > index_day_price['이평선_5']) | \
                                        (index_day_price['종가'] > index_day_price['이평선_10'])
    index_day_price = index_day_price.set_index('날짜')

    return day_price_dict, index_day_price


def Buke_S002_buy_list(dir, day):

    path = f"{dir}/trade_list/{day}"

    # 1. 거래일 가져오기
    trading_days_list = read_trading_days(dir, day)

    # 2. 일봉 가져오기
    day_price = read_day_price(dir, day)
    index_day_price = read_index_day_price(dir)

    # 3. 지수 종목 구성 내역 가져오기
    kospi_history, kosdaq_history = read_index_history(dir, day)

    # 4. 매수 조건 및 우선순위 생성
    day_price, index_day_price = add_indicator(day_price, index_day_price)

    # 5. 매수할 종목 확인
    kospi_ticker_list = kospi_history['종목리스트'].loc[trading_days_list[-1]].split(',')
    kosdaq_ticker_list = kosdaq_history['종목리스트'].loc[trading_days_list[-1]].split(',')

    # Market Timing
    try:
        yesterday_market_time = bool(index_day_price['#market_timing'].loc[trading_days_list[-1]])
        print(yesterday_market_time)

    except:

        print("오늘은 매매하기 위험한 날 입니다. 프로그램을 종료합니다.")
        none_list = ["오늘은 매매하기 위험한 날 입니다."]
        df = pd.DataFrame(none_list)
        df.to_csv(f"{path}/buy_2_{str(day)[-5:]}.txt")
        return False

    if yesterday_market_time is False:
        print("오늘은 매매하기 위험한 날 입니다. 프로그램을 종료합니다.")
        none_list = ["오늘은 매매하기 위험한 날 입니다."]
        df = pd.DataFrame(none_list)
        df.to_csv(f"{path}/buy_2_{str(day)[-5:]}.txt")
        return False
    print("종목을 선정중입니다.\n")

    pre_buy_list = []
    for ticker in kospi_ticker_list + kosdaq_ticker_list:

        try:
            yesterday_buy_condition = bool(day_price[ticker]['#final_result'].loc[trading_days_list[-1]])
            if yesterday_buy_condition:
                priority_value = float(day_price[ticker]['#priority'].loc[trading_days_list[-1]])
                ticker_data = (priority_value, ticker)
                pre_buy_list.append(ticker_data)
        except:
            continue

    pre_buy_list.sort(reverse=True)

    pre_buy_list = pre_buy_list[:10]

    basket = {'종목코드': [],'종목명':[], '매수가격1': []}

    for priority_value, ticker in pre_buy_list:

        try:
            yesterday_low = int(day_price[ticker]['저가'].loc[trading_days_list[-1]])
            ticker_name = day_price[ticker]['종목명'].loc[trading_days_list[-1]]

            if ticker in kospi_ticker_list:
                buy_price1 = get_ask_price(yesterday_low * 0.98, Market.kospi)

            else:
                buy_price1 = get_ask_price(yesterday_low * 0.98, Market.kosdaq)


            basket['종목코드'].append(ticker)
            basket['종목명'].append(ticker_name)
            basket['매수가격1'].append(f"{buy_price1}원")

        except:
            continue

    df = pd.DataFrame(basket)
    df.to_csv(f"{path}/buy_2_{str(day)[-5:]}.txt")